--  htables.e : hash tables : version 1.00
--  jiri babor
--  jbabor@paradise.net.nz
--  03-Jan-04

include machine.e

sequence s,t
global integer  index1, index2          -- bucket index, collision index 

function shuffle(sequence s)
    -- random sequence shuffle

    integer  j,temp

    for i = length(s) to 2 by -1 do
        j = rand(i)
        temp = s[j]
        s[j] = s[i]
        s[i] = temp
    end for

    return s
end function


----------------------------------------------------------------
-- 32-bit left rotation : modified Tommy Carlie's code - thanks!
----------------------------------------------------------------

constant rotate_left = allocate(13), rotate_left_param = rotate_left + 1

poke(rotate_left, {         
    #B8,#00,#00,#00,#00,                -- 0: mov eax, dword param (2)
    #D1,#C0,                            -- 5: rol eax, 1
    #A3,#00,#00,#00,#00,                -- 7: mov [@param]
    #C3})                               -- C: ret
poke4(rotate_left+8,rotate_left_param)  -- @param

function rol(atom x)
    -- rotate 32-bit integer left
    poke4(rotate_left_param, x)
    call(rotate_left)
    return peek4s(rotate_left_param)
end function


----------------------------------------------------------------
-- create an array of 256 32-bit random integer such that any 
-- bit column contains exactly 128 zeros and 128 ones
----------------------------------------------------------------

s = repeat(0,32)                        
t = repeat(0,128) & repeat(1,128)
set_rand(24341)                        
for i = 1 to 32 do s[i] = shuffle(t) end for

-- exchange rows and columns : turn sequence s 90 degrees clockwise
t = repeat(repeat(0,32),256)            -- init output sequence 
for r=1 to 32 do
    for c=1 to 256 do
        t[c][32-r+1]=s[r][c]
    end for
end for

-- turn bit sequences into 32-bit integers
for i = 1 to 256 do t[i] = bits_to_int(t[i]) end for


----------------------------------------------------------------
-- hash function is adaptation of 'BUZhash' (author not given)
----------------------------------------------------------------

function hash(sequence key)
    atom h
    integer n

    n = length(key)
    h = t[n]
    for i = 1 to n do
        h = xor_bits(rol(h),t[key[i]]) 
    end for
    return h
end function


----------------------------------------------------------------
-- table routines
----------------------------------------------------------------

global function new(integer number_of_buckets)
    -- initialize new table
    -- number of buckets *must* be integral power of 2: 2,4,8,16..2**32
    -- minimum recommended number: 8

    if and_bits(number_of_buckets,number_of_buckets-1) then
        puts(1, "'new' error: entered number of buckets is not integral power of 2. Aborting...\n")
        abort(1)
    end if
    return repeat({{}}, number_of_buckets)
end function

global function exists(sequence table,sequence key)
    -- boolean: return index2: zero (false) if key not found

    sequence t

    index1 = and_bits(hash(key),length(table)-1)+1
    t = table[index1]
    index2 = find(key, t[length(t)])
    return index2
end function

global function fetch(sequence table,sequence key)
    -- return value associated with field

    sequence t

    index1 = and_bits(hash(key),length(table)-1)+1
    t = table[index1]
    index2 = find(key, t[length(t)])

    if not index2 then
        puts(1, "'fetch' error: \""&key&"\" key not found. Aborting...\n")
        abort(1)
    end if
    return t[index2]
end function

global function make(sequence table,sequence key,object value)
    -- create new field for new value

    sequence t

    index1 = and_bits(hash(key),length(table)-1)+1
    t = table[index1]
    index2 = length(t)
    t = append(t,append(t[index2],key))
    t[index2] = value
    table[index1] = t

    return table
end function

global function set(sequence table,sequence key,object new_value)
    -- replace existing value with new value

    sequence t

    index1 = and_bits(hash(key),length(table)-1)+1
    t = table[index1]
    index2 = find(key, t[length(t)])

    if not index2 then
        puts(1, "'set' error: \""&key&"\" key not found. Aborting...\n")
        abort(1)
    end if
    t[index2] = new_value
    table[index1] = t

    return table
end function

global function delete(sequence table,sequence key)
    --  remove entry with specified key from table

    sequence t
    integer n

    index1 = and_bits(hash(key),length(table)-1)+1
    t = table[index1]
    n = length(t)
    index2 = find(key, t[n]) 
    if index2 then
        t[n] = t[n][1..index2-1] & t[n][index2+1..n-1]
        t = t[1..index2-1] & t[index2+1..n]
    else
        puts(1,"'delete' error: \""&key&"\" key not found. Aborting...\n")
        abort(1)
    end if
    table[index1] = t

    return table
end function
